/* this is a simple program to exercise the ipc mechanisms defined in
   ipc.c. use -DDEBUG when you compile in order to get detailed error
   messages.

   Compiling:
     Define SOLARIS2 or SUNOS4 as appropriate.
     If you are compiling under SOLARIS2 you will also need the 
     posix 4 library (-lposix4).
*/
#ifdef SUNOS4
#ifndef __USE_FIXED_PROTOTYPES__
#define __USE_FIXED_PROTOTYPES__
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>

#ifdef SOLARIS2
#include <time.h>
#endif

#include "ipc.h"

#ifdef SUNOS4
extern void srand48(int);
extern double drand48(void);
extern int usleep(unsigned useconds);
#endif

#ifdef SOLARIS2
int usleep(unsigned useconds);
#endif

//who is the crazy who put 5000 at the beginning???
#define REPS 150

int gettime(int min, int max);
int testshm(int n, int m);
int testsems(int reps);

int main(int argc, char **argv)
{
  int n=5,m=10;
  int i=0;

  if (argc>1)  {
    n=atoi(argv[1]);
  }
  if (argc>2) {
    m=atoi(argv[2]);
  }

  fprintf(stderr,"testing shared memory:\n");
  i=testshm(n,m);
  fprintf(stderr,"testshm: there %s %d error%s\n",(i==1?"was":"were"),i,(i==1?"":"s"));

  fprintf(stderr,"\ntesting semaphores:\n");
  i=testsems(REPS);
  fprintf(stderr,"testsems: there %s %d error%s\n",(i==1?"was":"were"),i,(i==1?"":"s"));

  return 0;
}

int testshm(int n, int m)
{
  int i,j;
  int **array;
  int errors=0;

  array=malloc(n*sizeof(*array));

  fprintf(stderr,"  num arrays: %d, array size: %d\n",n,m);

  for (i=0; i<n; i++) {
    if ((array[i]=shmalloc(m*(sizeof(**array)))) == NULL) {
      perror("  shmalloc");
      exit(1);
    }
    fprintf(stderr,"  allocated %d arrays\r",n);
  }
  fprintf(stderr,"  allocated %d arrays\n",n);
  sleep(5);

  switch (fork()) {
  case -1:
    perror("  fork");
    break;
  case 0:
    /* child */
    for (i=0; i<n; i++) {
      for (j=0; j<m; j++) {
	fprintf(stderr,"  child initializing array[%d][%d]\r",i,j);
	array[i][j]=i*j;
      }
    }

    fprintf(stderr,"  child initialized %d arrays         \n",i);
    exit(0);
    break;
  default:
    /* wait for child to finish */
    wait(NULL);
  }


  fprintf(stderr,"  parent will now test %d arrays       \n",n);
  
  for (i=0; i<n; i++) {
    for (j=0; j<m; j++) {
      fprintf(stderr,"  parent testing array[%d][%d]\r",i,j);
      if (array[i][j] != (i*j)) {
	fprintf(stderr,"  array[%d][%d] != %d\n",i,j,i*j);
	errors++;
      }
    }
  }

  fprintf(stderr,"  parent tested %d arrays       \n",n);

  sleep(5);

  fprintf(stderr,"  parent will now free arrays\n");

  /* free in different order than we allocated */
  /* even numbers first */
  for (i=0; i<n; i+=2) {
    if (shfree(array[i]) == -1) {
      perror("  shfree");
      errors++;
    }
  }
  /* then odd numbers */
  for (i=1; i<n; i+=2) {
    if (shfree(array[i]) == -1) {
      perror("  shfree");
      errors++;
    }
  }
  fprintf(stderr,"  freed %d arrays\n",n);

  return errors;
}

#define MAX_WAIT 5
#define MAX(a,b) (a>b?a:b)
#define MIN(a,b) (a>b?b:a)

int testsems(int reps)
{
  double max=0.0;
  double min=0.0;
  semid_t sem;
  double *nums;
  int i;
  double d;
  int errors=0;

  if ((nums=shmalloc(2*sizeof(*nums))) == NULL) {
    fprintf(stderr,"  shmalloc failed\n");
    return 1;
  }

  srand48(getpid());
  nums[0]=0.0;
  nums[1]=1.0;
  if ((sem=semcreate(1)) == -1) {
    perror("  semcreate");
    fprintf(stderr,"  couldn't create semaphore\n");
    return 1;
  }

  fprintf(stderr,"  exercising semaphore (this will take some time)...\n");
  fprintf(stderr,"  2 processes will modify array in critical section\n");
  fprintf(stderr,"  initial values: nums[0]=%f, nums[1]=%f\n",nums[0],nums[1]);

  switch(fork()) {
  case -1:
    perror("  fork");
    return 1;
  case 0:
    for (i=0; i<reps; i++) {
      usleep(gettime(0,MAX_WAIT));

      /* decrease nums by their difference (should be one) */
      p(sem);
      d=nums[1]-nums[0];
      if (d-1.0 > 0.001) {
	fprintf(stderr,"  diff = %f\n",d);
	fprintf(stderr,"  nums[0]=%f, nums[1]=%f\n",nums[0],nums[1]);
	errors++;
	break;
      }
      else {
	nums[0]-=d;
	usleep(gettime(0,MAX_WAIT));
	nums[1]-=d;
	v(sem);
      }
    }
    exit(1);
  default:
    for (i=0; i<reps; i++) {
      usleep(gettime(0,MAX_WAIT));

      /* increase nums by their difference (should be one) */
      p(sem);
      min=MIN(min,nums[0]);
      min=MIN(min,nums[1]);

      d=nums[1]-nums[0];
      if (d-1.0 > 0.001) {
	fprintf(stderr,"  diff = %f\n",d);
	fprintf(stderr,"  nums[0]=%f, nums[1]=%f\n",nums[0],nums[1]);
	errors++;
	break;
      }
      else {
	nums[0]+=d;
	usleep(gettime(0,MAX_WAIT));
	nums[1]+=d;
	max=MAX(max,nums[0]);
	max=MAX(max,nums[1]);
	v(sem);
      }
    }
    wait(NULL);
  }

  fprintf(stderr,"  min: %f, max: %f\n",min,max);
  fprintf(stderr,"  values should differ by 1.0:\n");
  fprintf(stderr,"  final values: nums[0]=%f, nums[1]=%f, diff=%f\n",
	  nums[0],nums[1],nums[1]-nums[0]);
  d=nums[1]-nums[0];
  if (d < 1.0-0.001 || d > 1.0+0.001) {
    fprintf(stderr,"  diff greater than 1.0\n");
    errors++;
  }

  if (semdestroy(sem) == -1) {
    perror("  semdestroy");
    fprintf(stderr,"  couldn't destroy semaphore\n");
    errors++;
  }

  shfree(nums);

  return errors;
}

int gettime(int min, int max)
{
  return (drand48()*(max-min+1))+min;
}

#ifdef SOLARIS2
int usleep(unsigned useconds)
{
  static struct timespec nseconds;
  static struct timespec rseconds;

  nseconds.tv_nsec=useconds%1000000;
  nseconds.tv_sec=(useconds-nseconds.tv_nsec)/1000000;

  return (nanosleep(&nseconds,&rseconds));
}
#endif
