/* this program checks that the SEM_UNDO feature works correctly.

   1) the child grabs the semaphore, 
   2) the parent blocks on the semaphore
   3) the child exits without releasing the semaphore
   4) what happens now depepnds on the value of the flags passed to
   semop() by semwait() and semsignal(): If SEM_UNDO is set, the
   parent should now unblock and get the semaphore. If SEM_UNDO is not
   set, the parent will continue to block until the child (which no
   longer exists) explicitly releases the semaphore. i.e the parent is
   deadlocked.

   See the manual pages for exit(2) and semop(2).
*/
   
#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <unistd.h>
#include "ipc.h"

int main(int argc, char **argv)
{
  semid_t s;
  int i;

  s=semcreate(1);

  switch (fork()) {
  case -1:
    perror("fork");
    break;
  case 0:
    fprintf(stderr,"child: p(s)\n");
    p(s);
    fprintf(stderr,"child: got sem\n");
    fprintf(stderr,"child: sleeping 5 sec\n");
    sleep(5);
    fprintf(stderr,"child: exiting\n");
    exit(0);
  default:
    fprintf(stderr,"parent: sleeping 2 sec\n");
    sleep(2);
    fprintf(stderr,"parent: p(s)\n");
    if ((i=p(s)) == -1) {
      perror("semwait");
    }
    fprintf(stderr,"parent: got sem (%d)\n",i);

    /* clean up */
    wait(NULL);
  }

  semdestroy(s);
  return 0;
}
