/**
 * This is a skeleton for the cracker program.
 *
 * Your work is to complete the main function
 * and to understand the rest
 */

#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<sys/types.h>
#include<sys/wait.h>
#include<signal.h>
#include<string.h>
#include "guess.h"

/** this is useful to READ the program **/
#define READ  (thePipe[0])
#define WRITE (thePipe[1])

/** size for the buffer ; the words have a smaller size! **/
#define BUFFER_SIZE 256

/**
 * declaration of the processes
 */
void producer(int fd,char* filename);
void consumer(int fd);

/**
 * the main program: creates the pipe
 * and starts the producer and the consumer
 */
int main(int argc,char *argv[]) {

  //UNCOMMENT THE NEXT LINE FOR THE ASSIGNMENT!
  //int thePipe[2];       /* stores the descriptor pair */
  pid_t pid1,pid2,pid;  /* process IDs */
  int status;           /* used to collect wait statuses */

  /* checking arguments */
  if (argc<2) {
    fprintf(stderr,"File name for the words needed!\n");
    fflush(stderr);
    exit(1);
  }

  /*****************
   ** HERE YOU WORK!
   *****************/

  //pipe creation, 2*fork, be strict with error
  //handling, and do not forget to close useless
  //pipes
  //First process MUST have pid1 and second one pid2





  // you should not modify the rest of the program

  /** now we wait for the 2 children which are working **/
  /** and we do this cleanly to be sure that every process is
   ** dead when the program exits **/

  pid=wait(&status);
  if (WIFEXITED(status)) {
      printf("status: exit(%d)\n",
             WEXITSTATUS(status));
      fflush(stdout);
  }

  /* if abnormal termination */
  if (!(WIFEXITED(status) && WEXITSTATUS(status)==0)) {
    /* kill the other process */
    if (pid==pid1)
      kill(pid2,2);
    else if (pid==pid2)
      kill(pid1,2);
    /* and wait for the second one */
    wait(NULL);
    fprintf(stderr,"Error, program stopped!\n");
    fflush(stderr);
    exit(1);
  }

  printf("Process %d has finished!\n",(int)pid);
  fflush(stdout);
  /* normal wait for the second process */
  pid=wait(NULL);
  printf("Process %d has finished!\n",(int)pid);
  fflush(stdout);

  return 0;
}

/**
 * producer process ; DO NOT CHANGE THIS
 */
void producer(int pipeWriteSide,char* filename) {
  /* buffers */
  static char buffer[BUFFER_SIZE];
  static char word[BUFFER_SIZE];
  /* stream holder */
  FILE *instream;
  /* the counter to count the words: check what you send is received */
  int counter=0;
  /* you may add variable declaration here, remember: you're in C
   * so the declarations are to be done here */

  /* catch pipe signals: which signals? Well a SIGTERM signal is sent
   * if a process tries to write to a pipe where no process is connected
   * at the other end: no possible read. */
  signal(SIGPIPE,SIG_IGN);

  /* open the file for reading */
  instream=fopen(filename,"r");
  if (instream==NULL) {
    fprintf(stderr,"Can't open %s\n",filename);
    fflush(stderr);
    /* close the pipe in case of error */
    close(pipeWriteSide);
    /* and error code */
    exit(1);
  }

  /* read words separated by space, new lines... */
  while(fgets(buffer,BUFFER_SIZE,instream)) {
    /* while something read, send it to the consumer */
    /* if there was a match of one word */
    if (sscanf(buffer,"%[a-zA-Z1-9./]",word)==1) {
      int len=strlen(word)+1; /* +1: +\0 */
      /* send the size of the word */
      if(write(pipeWriteSide,(char*)&len,sizeof(int)) != sizeof(int)) 
          break;
      /* ant the word itself */
      if(write(pipeWriteSide,word,len) != len)
  	  break;
      /* count the words */
      counter++;
    }
  }

  /* print the counter */
  printf("Words sent: %d words\n",counter);
}


/**
 * consumer process ; DO NOT CHANGE THIS
 */
void consumer(int pipeReadSide) {
  /* buffer */
  static char buffer[BUFFER_SIZE];
  int bytes,counter,len,remaining,i;
  /* flags for the cracked accounts */
  int cracked[numaccts];

  //initialization
  for(i=0;i<numaccts;i++)
    cracked[i]=0;
  remaining=numaccts;
  counter=0;
  printf("%d accounts to crack, let's go!\n",remaining);
  fflush(stdout);
  
  do {
    /* word length first */
    bytes=read(pipeReadSide,(char*)&len,sizeof(int));

    if (bytes>0) {

      if (bytes!=sizeof(int)) {
	if (bytes<0)
	  perror("read");
	else {
	  fprintf(stderr,"read(int) failed (%d read)\n",bytes);
	  fflush(stderr);
	}
	/* close the pipe */
        close(pipeReadSide);
        /* error */
        exit(1);
      }

      /* then the word */
      bytes=read(pipeReadSide,buffer,len);

      /* if something read */
      if (bytes>0) {
        counter++;
        /* printf("Try %s\n",buffer); */
        if (counter%500==0) {
          printf("%d words tried so far...\n",counter);
          fflush(stdout);
        }
        for(i=0;i<numaccts;i++)
          /* if not yet cracked then crack now */
          if (!cracked[i] && guess(accounts[i],buffer)) {
            cracked[i]=1;
            remaining--;
            printf("Account %s cracked, password=%s\n",
                   accounts[i],buffer);
            fflush(stdout);
          }
        
      } // if (bytes>0)
      /* else if error */
      else if (bytes<0) {
        perror("read");
        /* close the pipe */
        close(pipeReadSide);
        /* and error code */
        exit(1);
      }
    } // if (bytes>0)

  } while(bytes!=0 && remaining>0);

  /* print the results */
  printf("%d accounts cracked, %d not cracked, %d words tried\n",
         numaccts-remaining,remaining,counter);
  fflush(stdout);

}



